/* redirect.c -- send back a redirect for a chosen URL

	Copyright 1999,2005,2006,2007,2008
		by Mark E. Mallett, MV Communications, Inc.

	See the package's "LICENSE" file for terms.

Used as the target of an HTML form to pass back a redirected
URL reference to the browser.  The form passes the URL in the
"redirect-url" tag and this CGI program emits a response directing
the client browser to that new URL.

*/


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <net/if.h>
#include <netdb.h>
#ifndef SIOCGIFCONF
# include <sys/sockio.h>
#endif

#include "mvform.h"

#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-in.h>
#include <mml/mml-iopf.h>
#include <mml/mml-cgi.h>

#include "formmisc.h"

/* Local Definitions */


/* External data referenced */


/* External routines used */

extern	char	*getenv();

/* Local data publicly available */

    /* These are not used; they need to exist to satisfy linker references
       because we use some non-related function in an external module.
    */
	char	*DocumentRootP;		/* Document root */
	char	Formdir[PATH_MAX +1];	/* Dir for this form */

/* Local routines and forward references */

    /* Add an IP(v4) address to the local IP addresses list */
static	BOOL	add_local_ip_addr PROTO( (in_addr_t addr) );

    /* Get local IP addresses on this host */
static	BOOL	get_local_ip_addrs PROTO( (void) );

    /* Check the referrer */
static	BOOL	referrer_ok PROTO( (char *refP) );

    /* Check the target */
static	BOOL	target_ok PROTO( (char *urlP) );


/* Private data */

static	int	MyipC;			/* Count of addresses in MyipP */
static	int	MyipL;			/* Length of MyipP array */
static	in_addr_t *MyipP;		/* IPv4 addresses on this host */

    /* A list of valid referrer prefix strings.  These will be checked
       in addition to the default IP address check.  NULL ends the list.
    */
static	char	*Referrers_ok[] = {
	"http://invalid/",

	NULL	
};


    /* A list of valid target prefix strings.  These will be checked
       in addition to the default IP address check.  NULL ends the list.
    */
static	char	*Targets_ok[] = {
	"http://invalid/",

	NULL	
};


main( argc, argv )
	int		argc;
	char		**argv;
{
	char		*errP;
	char		*newurlP;	/* Ptr to the new URL */
	char		*qP;		/* Ptr to qarg */
	char		*strP;

    /* put stderr onto stdout */
    close( 2 );
    dup2( 1, 2 );

    /* Get local IP addresses on this system */
    if ( !get_local_ip_addrs() ) {
	report_error( "get_local_ip_addrs() failed." );
	exit( 0 );
    }

    /* Collect arguments */
    if ( ( errP = qargs_in( (char *)NULL ) ) != NULL ) {
	report_error( errP );
	exit( 0 );
    }

    /* Find the redirect-url tag */
    if ( ( qP = qarg_find_tag( "redirect-url" ) ) == NULL ) {
	report_error(
	"The redirect-url tag was not properly passed by the invoking form." );
	exit( 0 );
    }

    newurlP = qarg_value( qP );		/* The new URL */


    /* We don't want to do redirects from one third party to another.  If
       the target is local or otherwise OK, we'll allow it, otherwise
       we'll check the referrer to see that it's local.
    */
    if ( !target_ok( newurlP ) ) {
	strP = getenv( "HTTP_REFERER" );
	if ( !referrer_ok( strP ) ) {
	    report_error( "Invalid referrer \"%s\"",
		          strP == NULL ? "" : strP );
	    exit( 0 );
	}
    }

    /* Emit the rest of the header, with the new location ref. */
    printf( "Location: %s\n", newurlP );
    printf( "Status: 302\n" );
    printf( "Connection: close\n" );
    begin_doc_out();


    /* Give a document in case the redirection didn't work. */
#if 1
    printf( "<title> Redirection to %s </title>\n", newurlP );
    printf( "<h1> Redirection to %s </h1>\n", newurlP );
    printf( "You are now requested to go to <br>\n" );
#endif
    printf( "<a href=\"%s\"> %s </a>\n", newurlP, newurlP );

    fflush( stdout );
}

/***********************************************************************
 *                                                                     *
 *                     Internal support routines                       *
 *                                                                     *
 ***********************************************************************/

/*

*//* add_local_ip_addr( addr )

	Adds an IPv4 address to the list of local IP addresses

Accepts :

	addr		IP address to add

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Does not add duplicate addresses (returns OK, tho)

	Stolen from mvmf's mvmtr program.

*/

static BOOL
add_local_ip_addr ARGLIST( (addr) )
    FARG( in_addr_t	addr )		/* The address to add */
{
	int		i;

    /* Don't add duplicate */
    for ( i = 0; i < MyipC; ++i )
	if ( MyipP[i] == addr )
	    return ( TRUE );

    /* Expand the list if necessary */
    if ( i == MyipL ) {
	/* Need to expand the array */
	i = MyipL + 10;			/* New size to aim for */
	if ( MyipL == 0 )
	    MyipP = (in_addr_t *)emalloc( "add_local_ip_addr", "new array",
					  i * sizeof( in_addr_t ) );
	else
	    MyipP = (in_addr_t *)erealloc( "add_local_ip_addr", "array",
					    MyipP,
					    i * sizeof( in_addr_t ) );
	if ( MyipP == NULL ) {
	    MyipL = MyipC = 0;
	    return ( FALSE );
	}

	MyipL = i;
    }

    /* Add new entry */
    MyipP[ MyipC++ ] = addr;

    return ( TRUE );
}
/*

*//* get_local_ip_addrs()

	Get the IP addresses associated with this system

Accepts :


Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Does a best effort attempt.  Failure to get addresses will still
	  give a TRUE return.  FALSE is reserved for real errors like
	  problems allocating memory.

	Stores results in global values.

	Stolen from mvmf's mvmtr program.

*/

static BOOL
get_local_ip_addrs NOARGLIST
{
	int		ifX;		/* Interface index in some form */
	int		len;
	int		sockfd;
	char		*ifnameP;
	in_addr_t	addr;
	struct ifreq	ifreq;
	long		ifdataL;
	UBYTE		*ifdataP;	/* Where ifreq data is stored */
	struct ifconf	ifconf;
	struct ifreq	*ifreqP;	/* working ifreq ptr */

    /* Need a handy fd to use the ioctls on */
    sockfd = socket( AF_INET, SOCK_DGRAM, IPPROTO_UDP );
    if ( sockfd < 0 )
        return ( TRUE );

#ifdef	SIOCGSIZIFCONF
    /* Some systems support SIOCGSIZIFCONF ioctl that will tell the
       size needed for the ifreq array.  We'll use it as a starting
       point at least.
    */
    if ( ioctl( sockfd, SIOCGSIZIFCONF, &ifdataL ) < 0 )
#else	/* SIOCGSIZIFCONF -- funky else */
	ifdataL = 10 * sizeof(struct ifreq);
#endif	/* SIOCGSIZIFCONF */

    /* Get initial buffer for ifreqs */
    ifdataP = (UBYTE *)emalloc( "get_local_ip_addrs", "ifdataP", ifdataL );
    if ( ifdataP == NULL ) {
	close( sockfd );
	return ( FALSE );
    }

    /* Loop to try to get interface list and expand buffer needed */
    for ( ; ; ) {
	ifconf.ifc_len = ifdataL;
	ifconf.ifc_req = (struct ifreq *)ifdataP;
	if ( ioctl( sockfd, SIOCGIFCONF, &ifconf ) < 0 ) {
	    warning( "get_local_ip_addrs: ioctl SIOCGIFCONF failed." );
	    close( sockfd );
	    return ( FALSE );
	}

	/* This ioctl will return the length of the data actually returned.
	   If the returned length is fairly close to the buffer length,
	   such that there wasn't room to fit one more, we probably didn't
	   get them all.  Dumb.  At any rate, keep reallocating if we
	   think this could conceivably have happened.
	*/
	if ( ( ifdataL - ifconf.ifc_len ) > ( sizeof( struct ifreq ) + 100 ) )
	    break;

	/* Try to expand the buffer, maybe */
	if ( ifdataL >= ( 1000 * sizeof( struct ifreq ) ) ) {
	    /* Gotta stop sometime */
	    warning( "get_local_ip_addrs: truncating interface list." );
	    break;
	}
	ifdataL += ( 50 * sizeof( struct ifreq ) );
	ifdataP = (UBYTE *)erealloc( "get_local_ip_addrs", "ifdataP",
				     ifdataP, ifdataL );
	if ( ifdataP == NULL ) {
	    close( sockfd );
	    return ( FALSE );
	}
    }

    /* Look through the interface list */
    ifX = 0;				/* Offset into ifdataP */

    for ( ; ;  ) {
	/* How do we know if we are at the end?  If the returned length
	   from the SOICGIFCONF ioctl won't cover this ifreq index.
	*/
	if ( ( ifX + sizeof( struct ifreq ) ) > ifconf.ifc_len )
	    break;
	ifreqP = (struct ifreq *)&ifdataP[ifX];
	ifnameP = &ifreqP->ifr_name[0];

	/* Skip ifX to the next entry */
#ifdef	HAVE_SOCKADDR_SA_LEN
	len = sizeof( ifreqP->ifr_name ) + ifreqP->ifr_addr.sa_len;
	if ( len < sizeof( struct ifreq ) )
	    len = sizeof( struct ifreq );
#else	/* HAVE_SOCKADDR_SA_LEN */
	len = sizeof( struct ifreq );
#endif	/* HAVE_SOCKADDR_SA_LEN */
	ifX += len;

	/* Get flags to see if we should pay attention */
	strcpy( &ifreq.ifr_name[0], ifnameP );
	if ( ioctl( sockfd, SIOCGIFFLAGS, &ifreq ) < 0 )
	    continue;			/* No data, just ignore */
	if ( ( ifreq.ifr_flags & IFF_UP ) == 0 )
	    continue;			/* Skip down interfaces */

	/* Handle this entry. */
	switch( ifreqP->ifr_addr.sa_family ) {
	    case AF_INET:		/* IPv4 */
		addr = ((struct sockaddr_in *)&ifreqP->ifr_addr)->sin_addr.s_addr;
		add_local_ip_addr( addr );
		break;

	    default:			/* Something we don't handle */
		break;
	}
    }

    close( sockfd );

    return ( TRUE );
}
/*

*//* referrer_ok( refP )

	See if the referrer is OK

Accepts :

	refP		The referrer string (URL that referred to this script)

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Checks the referrer to see if it's allowed to originate a
	request to this cgi-bin.  The checks are:

	 - Referrer string must be present;
	 - Referrer is OK if its initial substring matches any of
	   the strings in the Referrers_ok[] table;
	 - Referrer is OK if it contains a valid host reference, and
	   that host reference matches an IP address that's assigned to
	   this system (on any interface).

	Otherwise, referrer is not OK.

*/

static BOOL
referrer_ok ARGLIST( refP )
    FARG( char		*refP )		/* Referrer */
{
	int		i;
	char		*strP, *streP;
	char		**haPP;		/* ptr into host address list */
	in_addr_t	taddr;		/* Temp addr for comparison */
	struct hostent	*hP;
	char		buf[PATH_MAX +1];

    /* Bad if no referrer at all */
    if ( ( refP == NULL ) || ( *refP == NUL ) )
	return ( FALSE );

    /* OK if the initial part of the referrer matches any of the
       strings in the Referrers_ok table.  Case-insensitive compare.
    */
    for ( i = 0; ( strP = Referrers_ok[i] ) != NULL; ++i )
	if ( strnicmp( refP, strP, strlen( strP ) ) == 0 )
	    return ( TRUE );

    /* For http schemes, do an IP lookup of the host inside the referrer
       string.  If that IP address is local to THIS system, we'll take it.
    */
    if ( strnicmp( refP, "http://", 7 ) == 0 )
	strP = refP + 7;
    else if ( strnicmp( refP, "https://", 8 ) == 0 )
	strP = refP + 8;
    else
	strP = NULL;
    if ( strP != NULL ) {
	/* Hostname ends at the first '/' or at the end of the string */
	streP = strchr( strP, '/' );
	if ( streP != NULL )
	    i = streP - strP;
	else
	    i = strlen( strP );
	if ( i < ( sizeof(buf)-1 ) ) {
	    memcpy( buf, strP, i );
	    buf[i] = NUL;

	    hP = gethostbyname( buf );
	    if ( hP != NULL ) {
		/* Hostname is valid.  Check every IP address that it
		   resolves to against every local IP address.  If
		   there's a match, referrer will be judged as OK.
		*/
		for ( haPP = hP->h_addr_list; *haPP != NULL; ++haPP ) {
		    taddr = *(in_addr_t *)(*haPP);
		    for ( i = 0; i < MyipC; ++i ) {
			if ( taddr == MyipP[i] ) {
			    /* Got a match, presumed ok. */
			    return ( TRUE );
			}
		    }
		}
	    }
	}
    }

    /* No check succeeded, fail. */
    return ( FALSE );
}
/*

*//* target_ok( urlP )

	See if the target URL is OK

Accepts :

	urlP		The target URL string

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Checks the target URL to see if it's local or otherwise OK.
	  A target URL is OK if:

	 - it's a local reference (no domain name in URL);
	 - its initial substring matches any of the strings in
	   the Targets_ok[] table;
	 - the domain in the URL contains a valid host reference,
	   that host reference matches an IP address that's assigned to
	   this system (on any interface).

	Otherwise, target is not OK.

	This function is pretty dumb in looking for a hostname in the
	  target URL.  Basically if there is a "://" sequence it's
	  assumed a hostname follows.  In the presence of really weird
	  URLs this can end up being more paranoid than needed,
	  but at least that errs in the right direction.  Construct
	  your targets better if they fail, or rewrite this.

*/

static BOOL
target_ok ARGLIST( urlP )
    FARG( char		*urlP )		/* Target */
{
	int		i;
	char		*strP, *streP;
	char		**haPP;		/* ptr into host address list */
	in_addr_t	taddr;		/* Temp addr for comparison */
	struct hostent	*hP;
	char		buf[PATH_MAX +1];

    /* Bad if no target at all */
    if ( ( urlP == NULL ) || ( *urlP == NUL ) )
	return ( FALSE );

    /* OK if the initial part of the referrer matches any of the
       strings in the Targets_ok table.  Case-insensitive compare.
    */
    for ( i = 0; ( strP = Targets_ok[i] ) != NULL; ++i )
	if ( strnicmp( urlP, strP, strlen( strP ) ) == 0 )
	    return ( TRUE );

    /* If there is no "://" found, we'll take that to be the start of the
       host part of the URL.  However if there is no "://", then we can
       assume this is a relative reference which should always be OK.
    */
    strP = strstr( urlP, "://" );
    if ( strP != NULL ) {
	/* Found a "://" - we'll take that to be the start of the host
	   name.
	*/

	strP += 3;			/* skip the "://" */

	/* Hostname ends at the first '/' or at the end of the string */
	streP = strchr( strP, '/' );
	if ( streP != NULL )
	    i = streP - strP;
	else
	    i = strlen( strP );
	if ( i < ( sizeof(buf)-1 ) ) {
	    memcpy( buf, strP, i );
	    buf[i] = NUL;

	    hP = gethostbyname( buf );
	    if ( hP != NULL ) {
		/* Hostname is valid.  Check every IP address that it
		   resolves to against every local IP address.  If
		   there's a match, referrer will be judged as OK.
		*/
		for ( haPP = hP->h_addr_list; *haPP != NULL; ++haPP ) {
		    taddr = *(in_addr_t *)(*haPP);
		    for ( i = 0; i < MyipC; ++i ) {
			if ( taddr == MyipP[i] ) {
			    /* Got a match, presumed ok. */
			    return ( TRUE );
			}
		    }
		}
	    }
	}

	/* We found a hostname but can't identify it as local, so fail. */
	return ( FALSE );
    }

    /* There was no host, so we'll assume a relative redirect, which
       should always be OK.
    */
    return ( TRUE );
}
